"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var jsdom_1 = require("@forbeslindesay/jsdom");
var eventChannel_1 = require("./eventChannel");
var Tab_1 = require("./Tab");
var Browser = (function () {
    function Browser(options) {
        var _this = this;
        this._currentTab = null;
        this._nextTabID = 0;
        this._tabsByID = new Map();
        this._tabsByName = new Map();
        this._localStorage = new Map();
        this.cookies = new jsdom_1.CookieJar();
        this.onInactive = eventChannel_1.default();
        this.onActive = eventChannel_1.default();
        this.onBeforeParse = eventChannel_1.default();
        this._onTabClose = function (tab) {
            _this._tabsByID.delete(tab.id);
            if (tab.name) {
                _this._tabsByName.delete(tab.name);
            }
            if (_this._currentTab === tab) {
                _this.open({
                    url: 'about:blank',
                });
            }
        };
        this.runScripts = options.runScripts;
        this.open({
            url: 'about:blank',
        });
    }
    Browser.prototype.findTab = function (nameOrID) {
        var tabByName = this._tabsByName.get(nameOrID);
        if (tabByName) {
            return tabByName;
        }
        var tabByID = this._tabsByID.get(nameOrID);
        if (tabByID) {
            return tabByID;
        }
        return null;
    };
    /**
     * Opens and returns a tab.  If an open window by the same name already exists,
     * opens this window in the same tab.  Omit name or use '_blank' to always open
     * a new tab.
     */
    Browser.prototype.open = function (options) {
        var named = options.name && this.findTab(options.name);
        if (named) {
            if (this._currentTab) {
                this.onInactive.emit(this._currentTab);
            }
            named.setLocation(options);
            this._currentTab = named;
            this.onActive.emit(named);
            return named;
        }
        var id = '' + (this._nextTabID++);
        var tab = new Tab_1.default(this, options, id, this._localStorage, this._onTabClose);
        this._tabsByID.set(id, tab);
        if (options.name) {
            this._tabsByName.set(options.name, tab);
        }
        if (this._currentTab) {
            this.onInactive.emit(this._currentTab);
        }
        this._currentTab = tab;
        this.onActive.emit(tab);
        return tab;
    };
    Object.defineProperty(Browser.prototype, "currentTab", {
        get: function () {
            return this._currentTab;
        },
        enumerable: true,
        configurable: true
    });
    Browser.prototype.dispose = function () {
    };
    return Browser;
}());
exports.default = Browser;
