"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var uuid = require("uuid/v1");
var Browser_1 = require("./Browser");
var WebdriverStatus_1 = require("./WebdriverStatus");
var tough_cookie_1 = require("tough-cookie");
var XPathSelectorError = {
    status: 32 /* InvalidSelector */,
    value: { message: 'Taxi Rank does not support XPath selectors.' },
};
function createResponse(request, value) {
    return { status: WebdriverStatus_1.SuccessStatus, sessionId: request.params.sessionId, value: value };
}
function isWebdriverResponse(value) {
    return value && typeof value === 'object' && typeof value.status === 'number';
}
function withCallback(fn) {
    return new Promise(function (resolve, reject) {
        fn(function (err, res) {
            if (err)
                return reject(err);
            else
                resolve(res);
        });
    });
}
var ElementStore = (function () {
    function ElementStore() {
        this.elements = new Map();
        this.elementIDs = new Map();
        this._nextIndex = 0;
    }
    ElementStore.prototype.storeElement = function (tab, element) {
        var oldID = this.elementIDs.get(element);
        if (oldID) {
            return { ELEMENT: oldID };
        }
        var id = '' + (this._nextIndex++);
        this.elements.set(id, element);
        this.elementIDs.set(element, id);
        return { ELEMENT: id };
    };
    ElementStore.prototype.getElement = function (reference) {
        return this.elements.get(typeof reference === 'string' ? reference : reference.ELEMENT) || null;
    };
    return ElementStore;
}());
var WebdriverSession = (function () {
    function WebdriverSession(browser) {
        this.elements = new ElementStore();
        this.timeouts = new Map([
            ["async" /* ASYNC_SCRIPT */, 5000],
            ["implicit" /* IMPLICIT */, 5000],
            ["page load" /* PAGE_LOAD */, 5000],
            ["script" /* SCRIPT */, 5000],
        ]);
        this.browser = browser;
    }
    return WebdriverSession;
}());
var Webdriver = (function () {
    function Webdriver() {
        this._activeSessions = new Map();
    }
    Webdriver.prototype.createSession = function (request) {
        return __awaiter(this, void 0, void 0, function () {
            var runScripts, capabilities, sessionId, currentSession;
            return __generator(this, function (_a) {
                runScripts = request.body.desiredCapabilities.runScripts || request.body.desiredCapabilities.runScripts || 'dangerously';
                if (runScripts !== 'dangerously' && runScripts !== 'outside-only') {
                    return [2 /*return*/, {
                            status: 33 /* SessionNotCreatedException */,
                            value: { message: 'The only valid values for the runScripts capability are "dangerously" and "outside-only"' },
                        }];
                }
                capabilities = { runScripts: runScripts };
                sessionId = uuid();
                currentSession = new WebdriverSession(new Browser_1.default({ runScripts: capabilities.runScripts }));
                this._activeSessions.set(sessionId, currentSession);
                // TODO: initialise lots of things from capabilities here
                return [2 /*return*/, { status: WebdriverStatus_1.SuccessStatus, sessionId: sessionId, value: capabilities }];
            });
        });
    };
    Webdriver.prototype.deleteSession = function (request) {
        return __awaiter(this, void 0, void 0, function () {
            var session;
            return __generator(this, function (_a) {
                session = this._activeSessions.get(request.params.sessionId);
                if (session) {
                    session.browser.dispose();
                    this._activeSessions.delete(request.params.sessionId);
                }
                return [2 /*return*/, createResponse(request, undefined)];
            });
        });
    };
    Webdriver.prototype._withSession = function (request, fn) {
        return __awaiter(this, void 0, void 0, function () {
            var session, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        session = this._activeSessions.get(request.params.sessionId);
                        if (!session) {
                            return [2 /*return*/, { status: 6 /* NoSuchSession */, value: { message: "No session with id, " + request.params.sessionId + " was found." } }];
                        }
                        return [4 /*yield*/, fn(session)];
                    case 1:
                        result = _a.sent();
                        if (isWebdriverResponse(result)) {
                            return [2 /*return*/, result];
                        }
                        else {
                            return [2 /*return*/, createResponse(request, result)];
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    Webdriver.prototype.setTimeouts = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                session.timeouts.set(request.body.type, request.body.ms);
                return [2 /*return*/];
            });
        }); });
    };
    Webdriver.prototype.setAsyncScriptTimeOut = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                session.timeouts.set("async" /* ASYNC_SCRIPT */, request.body.ms);
                return [2 /*return*/];
            });
        }); });
    };
    Webdriver.prototype.getActiveWindowHandle = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab;
            return __generator(this, function (_a) {
                tab = session.browser.currentTab;
                if (!tab) {
                    return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                }
                return [2 /*return*/, tab.id];
            });
        }); });
    };
    Webdriver.prototype.setUrl = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        dom.window.location.href = request.body.url;
                        return [4 /*yield*/, tab.whenReady()];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.goBack = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        dom.window.history.back();
                        return [4 /*yield*/, tab.whenReady()];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.goForward = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        dom.window.history.forward();
                        return [4 /*yield*/, tab.whenReady()];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.refresh = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        dom.window.location.reload();
                        return [4 /*yield*/, tab.whenReady()];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.getElementFromResponse = function (request, elementsResponse) {
        if (elementsResponse.status !== WebdriverStatus_1.SuccessStatus) {
            return elementsResponse;
        }
        var elements = elementsResponse.value;
        if (elements.length) {
            return createResponse(request, elements[0]);
        }
        return { status: 7 /* NoSuchElement */, value: { message: "There do not seem to be any elements with the selector \"" + request.body.value + "\"" } };
    };
    Webdriver.prototype.getElement = function (request) {
        return __awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        _a = this.getElementFromResponse;
                        _b = [request];
                        return [4 /*yield*/, this.getElements(request)];
                    case 1: return [2 /*return*/, _a.apply(this, _b.concat([_c.sent()]))];
                }
            });
        });
    };
    Webdriver.prototype._htmlCollectionToArray = function (session, tab, elements) {
        var result = [];
        for (var i = 0; i < elements.length; i++) {
            result.push(session.elements.storeElement(tab, elements[i]));
        }
        return result;
    };
    Webdriver.prototype._getElementsInContext = function (ctx) {
        function htmlCollectionToArray(elements) {
            var result = [];
            for (var i = 0; i < elements.length; i++) {
                result.push(elements[i]);
            }
            return result;
        }
        switch (ctx.using) {
            case "class name" /* CLASS */:
                return htmlCollectionToArray(ctx.parent.getElementsByClassName(ctx.value));
            case "css selector" /* CSS */:
                return htmlCollectionToArray(ctx.parent.querySelectorAll(ctx.value));
            case "id" /* ID */:
                return htmlCollectionToArray(ctx.parent.querySelectorAll('#' + ctx.value)).filter(function (element) { return element.id === ctx.value; });
            case "name" /* NAME */:
                return htmlCollectionToArray(ctx.parent.querySelectorAll("[name=\"" + ctx.value + "\"]")).filter(function (element) { return element.getAttribute('name') === ctx.value; });
            case "link text" /* LINK_TEXT */:
                return htmlCollectionToArray(ctx.parent.getElementsByTagName('a')).filter(function (element) { return element.textContent === ctx.value; });
            case "partial link text" /* PARTIAL_LINK_TEXT */:
                return htmlCollectionToArray(ctx.parent.getElementsByTagName('a')).filter(function (element) { return (element.textContent || '').indexOf(ctx.value) !== -1; });
            case "tag name" /* TAG */:
                return htmlCollectionToArray(ctx.parent.getElementsByTagName(ctx.value));
            case "xpath" /* XPATH */:
                throw new Error('XPath Selectors are not implemented');
        }
    };
    Webdriver.prototype.getElements = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, element;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, []];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        if (request.body.using === "id" /* ID */) {
                            element = dom.window.document.getElementById(request.body.value);
                            return [2 /*return*/, element ? [session.elements.storeElement(tab, element)] : []];
                        }
                        if (request.body.using === "xpath" /* XPATH */) {
                            return [2 /*return*/, XPathSelectorError];
                        }
                        return [2 /*return*/, this._getElementsInContext({
                                using: request.body.using,
                                value: request.body.value,
                                parent: dom.window.document,
                                session: session,
                            }).map(function (element) { return session.elements.storeElement(tab, element); })];
                }
            });
        }); });
    };
    Webdriver.prototype.getChildElement = function (request) {
        return __awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        _a = this.getElementFromResponse;
                        _b = [request];
                        return [4 /*yield*/, this.getChildElements(request)];
                    case 1: return [2 /*return*/, _a.apply(this, _b.concat([_c.sent()]))];
                }
            });
        });
    };
    Webdriver.prototype.getChildElements = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, []];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        if (request.body.using === "id" /* ID */) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'Cannot use an element ID selector on another element, only on body' } }];
                        }
                        if (request.body.using === "xpath" /* XPATH */) {
                            return [2 /*return*/, XPathSelectorError];
                        }
                        return [2 /*return*/, this._getElementsInContext({
                                using: request.body.using,
                                value: request.body.value,
                                parent: element,
                                session: session,
                            }).map(function (element) { return session.elements.storeElement(tab, element); })];
                }
            });
        }); });
    };
    Webdriver.prototype.compareElements = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, request.params.elementA === request.params.elementB];
            });
        }); });
    };
    Webdriver.prototype.getActiveElement = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, session.elements.storeElement(tab, dom.window.document.activeElement)];
                }
            });
        }); });
    };
    Webdriver.prototype.getSource = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, dom.serialize()];
                }
            });
        }); });
    };
    Webdriver.prototype.getTitle = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, dom.window.document.title];
                }
            });
        }); });
    };
    Webdriver.prototype.getUrl = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, dom.window.location.href];
                }
            });
        }); });
    };
    Webdriver.prototype.deleteAllCookies = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var store, cookies, _loop_1, _i, cookies_1, cookie;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        store = session.browser.cookies.store;
                        return [4 /*yield*/, withCallback(function (cb) { return store.getAllCookies(cb); })];
                    case 1:
                        cookies = _a.sent();
                        _loop_1 = function (cookie) {
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0: return [4 /*yield*/, withCallback(function (cb) { return store.removeCookie(cookie.domain, cookie.path, cookie.key, function (err) { return cb(err, undefined); }); })];
                                    case 1:
                                        _a.sent();
                                        return [2 /*return*/];
                                }
                            });
                        };
                        _i = 0, cookies_1 = cookies;
                        _a.label = 2;
                    case 2:
                        if (!(_i < cookies_1.length)) return [3 /*break*/, 5];
                        cookie = cookies_1[_i];
                        return [5 /*yield**/, _loop_1(cookie)];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.deleteCookie = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var store, cookies, _loop_2, _i, cookies_2, cookie;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        store = session.browser.cookies.store;
                        return [4 /*yield*/, withCallback(function (cb) { return store.getAllCookies(cb); })];
                    case 1:
                        cookies = _a.sent();
                        _loop_2 = function (cookie) {
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        if (!(cookie.key === request.params.key)) return [3 /*break*/, 2];
                                        return [4 /*yield*/, withCallback(function (cb) { return store.removeCookie(cookie.domain, cookie.path, cookie.key, function (err) { return cb(err, undefined); }); })];
                                    case 1:
                                        _a.sent();
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        };
                        _i = 0, cookies_2 = cookies;
                        _a.label = 2;
                    case 2:
                        if (!(_i < cookies_2.length)) return [3 /*break*/, 5];
                        cookie = cookies_2[_i];
                        return [5 /*yield**/, _loop_2(cookie)];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.setCookie = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [4 /*yield*/, new Promise(function (resolve, reject) {
                                var c = request.body.cookie;
                                session.browser.cookies.setCookie(new tough_cookie_1.Cookie({
                                    key: c.name,
                                    value: c.value,
                                    // expires: undefined,
                                    maxAge: c.expiry || 'Infinity',
                                    domain: c.domain,
                                    path: c.path,
                                    secure: c.secure || false,
                                    httpOnly: c.httpOnly || false,
                                    // extensions: [],
                                    creation: new Date(),
                                    // creationIndex: Date.now(),
                                    // hostOnly: false,
                                    // pathIsDefault: false,
                                    lastAccessed: new Date(),
                                }), dom.window.location.href, function (err) {
                                    if (err)
                                        reject(err);
                                    else
                                        resolve();
                                });
                            })];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.getCookies = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, cookies;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [4 /*yield*/, new Promise(function (resolve, reject) {
                                session.browser.cookies.getCookies(dom.window.location.href, function (err, cookies) {
                                    if (err)
                                        reject(err);
                                    else
                                        resolve(cookies);
                                });
                            })];
                    case 2:
                        cookies = _a.sent();
                        return [2 /*return*/, cookies.map(function (c) { return ({
                                name: c.key,
                                value: c.value,
                                path: c.path,
                                domain: c.domain,
                                httpOnly: c.httpOnly,
                                secure: c.secure,
                                expiry: typeof c.maxAge === 'number' ? c.maxAge : undefined,
                            }); })];
                }
            });
        }); });
    };
    Webdriver.prototype._withElement = function (request, fn) {
        return __awaiter(this, void 0, void 0, function () {
            var session, element, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        session = this._activeSessions.get(request.params.sessionId);
                        if (!session) {
                            return [2 /*return*/, { status: 6 /* NoSuchSession */, value: { message: "No session with id, " + request.params.sessionId + " was found." } }];
                        }
                        element = session.elements.getElement(request.params.elementId);
                        if (!element) {
                            return [2 /*return*/, { status: 7 /* NoSuchElement */, value: { message: "No element with id, " + request.params.elementId + " was found." } }];
                        }
                        return [4 /*yield*/, fn(session, element)];
                    case 1:
                        result = _a.sent();
                        if (isWebdriverResponse(result)) {
                            return [2 /*return*/, result];
                        }
                        else {
                            return [2 /*return*/, createResponse(request, result)];
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    Webdriver.prototype.getTagName = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, element.tagName.toLowerCase()];
            });
        }); });
    };
    Webdriver.prototype.getAttribute = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                if (request.params.attributeName === 'value' && isInputlike(element)) {
                    return [2 /*return*/, element.value];
                }
                if (request.params.attributeName === 'checked' && isInput(element)) {
                    // TODO: I'm not sure this is quite right
                    return [2 /*return*/, element.checked ? 'checked' : null];
                }
                return [2 /*return*/, element.getAttribute(request.params.attributeName)];
            });
        }); });
    };
    Webdriver.prototype.getCssProperty = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, dom.window.getComputedStyle(element)[request.params.propertyName] || null];
                }
            });
        }); });
    };
    Webdriver.prototype.getEnabled = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, !element.disabled];
            });
        }); });
    };
    Webdriver.prototype.getSelected = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, !!element.checked];
            });
        }); });
    };
    Webdriver.prototype.getText = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, element.textContent || ''];
            });
        }); });
    };
    Webdriver.prototype.click = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, style, top, left, x, y;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        style = dom.window.getComputedStyle(element);
                        top = style.top ? parseInt(style.top, 10) : 0;
                        left = style.left ? parseInt(style.left, 10) : 0;
                        x = style.width ? parseInt(style.width, 10) / 2 : 0;
                        y = style.height ? parseInt(style.height, 10) / 2 : 0;
                        ['mousedown', 'click', 'mouseup'].forEach(function (name) {
                            var e = dom.window.document.createEvent('MouseEvent');
                            e.initMouseEvent(name, true, true, dom.window, 0, left + x, top + y, left + x, top + y, false, false, false, false, 0 /* LEFT */, null);
                            element.dispatchEvent(e);
                        });
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.moveTo = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                if (!request.body.element) {
                    return [2 /*return*/, {
                            status: 13 /* UnknownError */,
                            value: { message: 'Taxi Rank does not support moving the mouse to arbitrary locations, you must specify an element.' },
                        }];
                }
                session.mouseLocation = {
                    elementID: request.body.element,
                    xoffset: request.body.xoffset || 0,
                    yoffset: request.body.yoffset || 0,
                };
                return [2 /*return*/];
            });
        }); });
    };
    Webdriver.prototype.globalClick = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var element, tab, dom, style, top, left, x, y;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!session.mouseLocation) {
                            return [2 /*return*/, {
                                    status: 13 /* UnknownError */,
                                    value: { message: 'Taxi Rank requires you to move the mouse before you can click it.' },
                                }];
                        }
                        element = session.elements.getElement(session.mouseLocation.elementID);
                        if (!element) {
                            return [2 /*return*/, {
                                    status: 13 /* UnknownError */,
                                    value: { message: 'Could not find an element with ID ' + session.mouseLocation.elementID + '.' },
                                }];
                        }
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        style = dom.window.getComputedStyle(element);
                        top = style.top ? parseInt(style.top, 10) : 0;
                        left = style.left ? parseInt(style.left, 10) : 0;
                        x = session.mouseLocation.xoffset;
                        y = session.mouseLocation.yoffset;
                        ['mousedown', 'click', 'mouseup'].forEach(function (name) {
                            var e = dom.window.document.createEvent('MouseEvent');
                            e.initMouseEvent(name, true, true, dom.window, 0, left + x, top + y, left + x, top + y, false, false, false, false, request.body.button, null);
                            element.dispatchEvent(e);
                        });
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.submit = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, form, submit;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        if (!isForm(element)) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'This element is not a form, it cannot be submitted.' } }];
                        }
                        form = element;
                        submit = dom.window.document.createEvent('HTMLEvents');
                        submit.initEvent('submit', true, true);
                        if (form.dispatchEvent(submit)) {
                            form.submit();
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 2:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.sendKeys = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, _i, _a, key, options, KeyboardEvent_1;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _b.sent();
                        for (_i = 0, _a = request.body.value; _i < _a.length; _i++) {
                            key = _a[_i];
                            options = {
                                key: key.toLowerCase(),
                                char: key,
                                charCode: key.charCodeAt(0),
                                shiftKey: key !== key.toLowerCase(),
                                keyCode: key.toUpperCase().charCodeAt(0),
                            };
                            KeyboardEvent_1 = dom.window.KeyboardEvent;
                            dom.window.document.activeElement.dispatchEvent(new KeyboardEvent_1('keydown', options));
                            dom.window.document.activeElement.dispatchEvent(new KeyboardEvent_1('keypress', options));
                            dom.window.document.activeElement.dispatchEvent(new KeyboardEvent_1('keyup', options));
                        }
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.execute = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, dom.window.Function('', request.body.script).apply(void 0, request.body.args)];
                }
            });
        }); });
    };
    Webdriver.prototype.executeAsync = function (request) {
        return this.execute(request);
    };
    Webdriver.prototype._getStorage = function (window, level) {
        switch (level) {
            case "local_storage" /* Local */:
                return window.localStorage;
            case "session_storage" /* Session */:
                return window.sessionStorage;
        }
    };
    Webdriver.prototype.setStorageItem = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        this._getStorage(dom.window, request.params.storageLevel).setItem(request.body.key, request.body.value);
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.getStorageItem = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, this._getStorage(dom.window, request.params.storageLevel).getItem(request.params.key)];
                }
            });
        }); });
    };
    Webdriver.prototype.getStorageKeys = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, storage, keys, i, key;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        storage = this._getStorage(dom.window, request.params.storageLevel);
                        keys = [];
                        for (i = 0; i < storage.length; i++) {
                            key = storage.key(i);
                            if (typeof key === 'string') {
                                keys.push(key);
                            }
                        }
                        return [2 /*return*/, keys];
                }
            });
        }); });
    };
    Webdriver.prototype.getStorageSize = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        return [2 /*return*/, this._getStorage(dom.window, request.params.storageLevel).length];
                }
            });
        }); });
    };
    Webdriver.prototype.clearStorage = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        this._getStorage(dom.window, request.params.storageLevel).clear();
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.removeStorageItem = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        this._getStorage(dom.window, request.params.storageLevel).removeItem(request.params.key);
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype._setValue = function (window, element, value) {
        if (!isInputlike(element)) {
            return { status: 12 /* InvalidElementState */, value: { message: 'You cannot set a value for an element with the tag name ' + element.tagName } };
        }
        if (element.disabled || element.readOnly) {
            return { status: 12 /* InvalidElementState */, value: { message: 'You cannot set a value for an element that is disabled or read only' } };
        }
        // Switch focus to field, change value and emit the input event (HTML5)
        element.focus();
        // `field.value = value` does not work if there is a custom property descriptor, e.g. in React
        // TODO: check what the actual spec compliant way of doing this is
        var descriptor = Object.getOwnPropertyDescriptor(element.constructor.prototype, 'value');
        if (!descriptor.set) {
            throw new Error('Corrupted element prototype');
        }
        descriptor.set.call(element, value);
        var e = window.document.createEvent('HTMLEvents');
        e.initEvent('input', true, true);
        element.dispatchEvent(e);
        // Switch focus out of field, if value changed, this will emit change event
        element.blur();
        return null;
    };
    Webdriver.prototype.clear = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        this._setValue(dom.window, element, '');
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.appendValue = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        if (!isInputlike(element)) {
                            return [2 /*return*/, { status: 12 /* InvalidElementState */, value: { message: 'You cannot set a value for an element with the tag name ' + element.tagName } }];
                        }
                        this._setValue(dom.window, element, element.value + request.body.value.join(''));
                        return [2 /*return*/];
                }
            });
        }); });
    };
    Webdriver.prototype.getDisplayed = function (request) {
        var _this = this;
        return this._withElement(request, function (session, element) { return __awaiter(_this, void 0, void 0, function () {
            var tab, dom, display;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tab = session.browser.currentTab;
                        if (!tab) {
                            return [2 /*return*/, { status: 13 /* UnknownError */, value: { message: 'No tab currently open' } }];
                        }
                        return [4 /*yield*/, tab.whenReady()];
                    case 1:
                        dom = _a.sent();
                        display = dom.window.getComputedStyle(element).display;
                        return [2 /*return*/, display !== 'none'];
                }
            });
        }); });
    };
    Webdriver.prototype.closeActiveWindow = function (request) {
        var _this = this;
        return this._withSession(request, function (session) { return __awaiter(_this, void 0, void 0, function () {
            var tab;
            return __generator(this, function (_a) {
                tab = session.browser.currentTab;
                if (tab) {
                    tab.close();
                }
                return [2 /*return*/];
            });
        }); });
    };
    return Webdriver;
}());
function isInput(element) {
    return element.tagName === 'INPUT';
}
function isInputlike(element) {
    return element.tagName === 'TEXTAREA' || element.tagName === 'INPUT';
}
function isForm(element) {
    return element.tagName === 'FORM';
}
exports.default = Webdriver;
