"use strict";
var express = require("express");
var body_parser_1 = require("body-parser");
var Webdriver_1 = require("./Webdriver");
var WebdriverUrlMap_1 = require("./WebdriverUrlMap");
function startServer(options) {
    if (options === void 0) { options = {}; }
    var app = express();
    app.use(body_parser_1.json());
    function isStorageLevel(str) {
        return str === "local_storage" /* Local */ || str === "session_storage" /* Session */;
    }
    var driver = new Webdriver_1.default();
    WebdriverUrlMap_1.default.forEach(function (pattern) {
        app[pattern.method](pattern.path, function (req, res, next) {
            if (req.params.storageLevel && !isStorageLevel(req.params.storageLevel)) {
                return next();
            }
            driver[pattern.fn](req).then(function (result) { return res.json(result); }, function (err) { return res.json({ status: 13 /* UnknownError */, value: { message: err.stack } }); }).catch(next);
        });
    });
    app.use(function (req, res, next) {
        res.json({ status: 9 /* UnknownCommand */, value: { message: "Unknown method: \"" + req.method + "\" \"" + req.path + "\"" } });
    });
    return app.listen(options.port || 9516, options.onStart);
}
;
module.exports = startServer;
